diff --git a/src/bin/psql/tab-complete.c b/src/bin/psql/tab-complete.c
index 3a43c09bf6..4ed241a4ee 100644
--- a/src/bin/psql/tab-complete.c
+++ b/src/bin/psql/tab-complete.c
@@ -976,6 +976,11 @@ static const SchemaQuery Query_for_list_of_statistics = {
 "       and pg_catalog.pg_table_is_visible(c2.oid)"\
 "       and c2.relispartition = 'true'"
 
+#define Query_for_list_of_cursors \
+" SELECT pg_catalog.quote_ident(name) "\
+"   FROM pg_catalog.pg_cursors "\
+"  WHERE substring(pg_catalog.quote_ident(name),1,%d)='%s'"
+
 /*
  * These object types were introduced later than our support cutoff of
  * server version 7.4.  We use the VersionedQuery infrastructure so that
@@ -2284,6 +2289,10 @@ psql_completion(const char *text, int start, int end)
 		COMPLETE_WITH_VERSIONED_SCHEMA_QUERY(Query_for_list_of_procedures, NULL);
 	else if (Matches("CALL", MatchAny))
 		COMPLETE_WITH("(");
+/* CLOSE */
+	else if (Matches("CLOSE"))
+		COMPLETE_WITH_QUERY(Query_for_list_of_cursors
+							" UNION SELECT 'ALL'");
 /* CLUSTER */
 	else if (Matches("CLUSTER"))
 		COMPLETE_WITH_SCHEMA_QUERY(Query_for_list_of_clusterables, "UNION SELECT 'VERBOSE'");
@@ -3005,8 +3014,43 @@ psql_completion(const char *text, int start, int end)
 	else if (Matches("DECLARE", MatchAny))
 		COMPLETE_WITH("BINARY", "INSENSITIVE", "SCROLL", "NO SCROLL",
 					  "CURSOR");
+	else if (Matches("DECLARE", MatchAny, "BINARY"))
+		COMPLETE_WITH("INSENSITIVE", "SCROLL", "NO SCROLL", "CURSOR");
+	else if (Matches("DECLARE", MatchAny, "INSENSITIVE"))
+		COMPLETE_WITH("BINARY", "SCROLL", "NO SCROLL", "CURSOR");
+	else if (Matches("DECLARE", MatchAny, "SCROLL") ||
+			 Matches("DECLARE", MatchAny, "NO", "SCROLL"))
+		COMPLETE_WITH("BINARY", "INSENSITIVE", "CURSOR");
+	else if (Matches("DECLARE", MatchAny, "BINARY", "INSENSITIVE") ||
+			 Matches("DECLARE", MatchAny, "INSENSITIVE", "BINARY"))
+		COMPLETE_WITH("SCROLL", "NO SCROLL", "CURSOR");
+	else if (Matches("DECLARE", MatchAny, "BINARY", "SCROLL") ||
+			 Matches("DECLARE", MatchAny, "BINARY", "NO", "SCROLL") ||
+			 Matches("DECLARE", MatchAny, "SCROLL", "BINARY") ||
+			 Matches("DECLARE", MatchAny, "NO", "SCROLL", "BINARY"))
+		COMPLETE_WITH("INSENSITIVE", "CURSOR");
+	else if (Matches("DECLARE", MatchAny, "INSENSITIVE", "SCROLL") ||
+			 Matches("DECLARE", MatchAny, "INSENSITIVE", "NO", "SCROLL") ||
+			 Matches("DECLARE", MatchAny, "SCROLL", "INSENSITIVE") ||
+			 Matches("DECLARE", MatchAny, "NO", "SCROLL", "INSENSITIVE"))
+		COMPLETE_WITH("BINARY", "CURSOR");
+	else if (Matches("DECLARE", MatchAny, "BINARY", "INSENSITIVE", "SCROLL") ||
+			 Matches("DECLARE", MatchAny, "BINARY", "INSENSITIVE", "NO", "SCROLL") ||
+			 Matches("DECLARE", MatchAny, "BINARY", "SCROLL", "INSENSITIVE") ||
+			 Matches("DECLARE", MatchAny, "BINARY", "NO", "SCROLL", "INSENSITIVE") ||
+			 Matches("DECLARE", MatchAny, "INSENSITIVE", "BINARY", "SCROLL") ||
+			 Matches("DECLARE", MatchAny, "INSENSITIVE", "BINARY", "NO", "SCROLL") ||
+			 Matches("DECLARE", MatchAny, "INSENSITIVE", "SCROLL", "BINARY") ||
+			 Matches("DECLARE", MatchAny, "INSENSITIVE", "NO", "SCROLL", "BINARY") ||
+			 Matches("DECLARE", MatchAny, "SCROLL", "BINARY", "INSENSITIVE") ||
+			 Matches("DECLARE", MatchAny, "NO", "SCROLL", "BINARY", "INSENSITIVE") ||
+			 Matches("DECLARE", MatchAny, "SCROLL", "INSENSITIVE", "BINARY") ||
+			 Matches("DECLARE", MatchAny, "NO", "SCROLL", "INSENSITIVE", "BINARY"))
+		COMPLETE_WITH("CURSOR");
 	else if (HeadMatches("DECLARE") && TailMatches("CURSOR"))
 		COMPLETE_WITH("WITH HOLD", "WITHOUT HOLD", "FOR");
+	else if (HeadMatches("DECLARE") && TailMatches("HOLD"))
+		COMPLETE_WITH("FOR");
 
 /* DELETE --- can be inside EXPLAIN, RULE, etc */
 	/* ... despite which, only complete DELETE with FROM at start of line */
@@ -3167,15 +3211,31 @@ psql_completion(const char *text, int start, int end)
 
 	/*
 	 * Complete FETCH with one of ABSOLUTE, BACKWARD, FORWARD, RELATIVE, ALL,
-	 * NEXT, PRIOR, FIRST, LAST
+	 * NEXT, PRIOR, FIRST, LAST, FROM, IN, and a list of cursors
 	 */
 	else if (Matches("FETCH|MOVE"))
-		COMPLETE_WITH("ABSOLUTE", "BACKWARD", "FORWARD", "RELATIVE",
-					  "ALL", "NEXT", "PRIOR", "FIRST", "LAST");
+		COMPLETE_WITH_QUERY(Query_for_list_of_cursors
+							" UNION SELECT 'ABSOLUTE'"
+							" UNION SELECT 'BACKWARD'"
+							" UNION SELECT 'FORWARD'"
+							" UNION SELECT 'RELATIVE'"
+							" UNION SELECT 'ALL'"
+							" UNION SELECT 'NEXT'"
+							" UNION SELECT 'PRIOR'"
+							" UNION SELECT 'FIRST'"
+							" UNION SELECT 'LAST'"
+							" UNION SELECT 'FROM'"
+							" UNION SELECT 'IN'");
 
-	/* Complete FETCH BACKWARD or FORWARD with one of ALL, FROM, IN */
+	/*
+	 * Complete FETCH BACKWARD or FORWARD with one of ALL, FROM, IN,
+	 * and a list of cursors
+	 */
 	else if (Matches("FETCH|MOVE", "BACKWARD|FORWARD"))
-		COMPLETE_WITH("ALL", "FROM", "IN");
+		COMPLETE_WITH_QUERY(Query_for_list_of_cursors
+							" UNION SELECT 'ALL'"
+							" UNION SELECT 'FROM'"
+							" UNION SELECT 'IN'");
 
 	/*
 	 * Complete FETCH <direction> with "FROM" or "IN". These are equivalent,
@@ -3185,7 +3245,13 @@ psql_completion(const char *text, int start, int end)
 	else if (Matches("FETCH|MOVE", "ABSOLUTE|BACKWARD|FORWARD|RELATIVE",
 					 MatchAnyExcept("FROM|IN")) ||
 			 Matches("FETCH|MOVE", "ALL|NEXT|PRIOR|FIRST|LAST"))
-		COMPLETE_WITH("FROM", "IN");
+		COMPLETE_WITH_QUERY(Query_for_list_of_cursors
+							" UNION SELECT 'FROM'"
+							" UNION SELECT 'IN'");
+	/* Complete FETCH <direction> "FROM" or "IN" with a list of cursors */
+	else if (HeadMatches("FETCH|MOVE") &&
+			 TailMatches("FROM|IN"))
+		COMPLETE_WITH_QUERY(Query_for_list_of_cursors);
 
 /* FOREIGN DATA WRAPPER */
 	/* applies in ALTER/DROP FDW and in CREATE SERVER */
